module Main where

import Auxiliary
import InputParser
import LPPE
import MLPPE
import ToPA
import ToAUT
import ToTrans
import Expressions
import ToPRISM
import System.IO
import System (getArgs)
import Simplify
import Confluence
import Sumelm
import Constelm
import Merging
import Parelm
import MaximalProgress
import DeadVariable
import Control.Monad
import FromPRISM
--import ToCTMC (let op: moet nog isMA toevoegen om weer compileerbaar te krijgen)
 
main = do
  args <- getArgs
  input <- hGetContents stdin

  -- Parsing the command-line parameters.
  let flags      = ["-conf", "-dead", "-mlppe", "-keeprates", "-maxprogress", "-fromprism", "-ma", "-pa", "-mapa", 
                    "-checkuntil", "-cadp", "-verbose", "-trans", "-lppe", "-size", "-prism", "-aut", "-visited", "-noprob",
                    "-ignorecycles", "-merge", "-statespace", "-prcrl", "-parelm", "-removecycles", "-PAstyle", "-strong", "-dtmc", "-nobasics"]

  let confluence           = elem "-conf" args                -- only for PA
  let dead                 = elem "-dead" args
  let merging              = elem "-merge" args
  let verbose              = elem "-verbose" args
  let printLPPE            = elem "-lppe" args                -- only for PA
  let printMLPPE           = elem "-mlppe" args               -- only for MA
  let mapa                 = elem "-mapa" args                -- only for MA
  let stateSpaceSize       = elem "-size" args
  let prism                = elem "-prism" args               -- only for PA
  let fromprism            = elem "-fromprism" args           -- only for PA
  let noprobs              = elem "-noprob" args
  let visited              = elem "-visited" args             -- only for PA
  let printAUTfile         = elem "-aut" args
  let keepRates            = elem "-keeprates" args           -- only for MA
  let dtmc                 = elem "-dtmc" args                -- only for PA
  let cadp                 = elem "-cadp" args            
  let states               = elem "-statespace" args    
  let ma                   = elem "-ma" args     
  let pa                   = elem "-pa" args || not(ma)
  let prcrl                = elem "-prcrl" args               -- only for PA
  let checkuntil           = elem "-checkuntil" args          -- only for PA
  let trans                = elem "-trans" args               -- only for PA
  let maxprogress          = elem "-maxprogress" args         -- only for MA
  let includingParelm      = elem "-parelm" args          
  let removecycles         = elem "-removecycles" args        -- only for PA
  let ignorecycles         = elem "-ignorecycles" args        -- only for PA
  let pastyle              = elem "-PAstyle" args || prcrl    -- do not use x := 5 notation in next state when pretty-printing (M)LPPE
  let strong               = elem "-strong" args              -- only for PA
  let nobasics             = elem "-nobasics" args
  let wrongFlags           = [x | x <- args, not(elem x flags || take 8 x == "-source=" || take 10 x == "-constant-")]
  let sourcefiles          = [x | ('-':'s':'o':'u':'r':'c':'e':'=':x) <- args]
  let constantdefs         = [x | ('-':'c':'o':'n':'s':'t':'a':'n':'t':'-':x) <- args]

  when (length wrongFlags > 0)          $ putStrLn ("Warning: flags " ++ show wrongFlags ++ " not recognised.")
  when (pa && ma)                       $ error ("Error: Choose either -pa or -ma.")
  when (not (pa || ma))                 $ error ("Error: Choose either -pa or -ma.")

  let paFlags                           = ["-conf", "-lppe", "-prism", "-fromprism", "-visited", "-dtmc", "-prcrl", "checkuntil", "-trans", "-removecycles", "-ignorecycles", "-strong"]
  let maFlags                           = ["-mlppe", "-mapa", "-keeprates", "-maxprogress"]

  let maErrors                          = [x | x <- args, elem x paFlags]
  let paErrors                          = [x | x <- args, elem x maFlags]

  when (ma && length maErrors > 0)      $ error ("Flags " ++ show maErrors ++ " not allowed in combination with -ma")
  when (pa && length paErrors > 0)      $ error ("Flags " ++ show paErrors ++ " not allowed in combination with -pa")

  -- Parsing the constants.
  let constants               = [(takeWhile (/= '=') constant, Variable (drop 1 (dropWhile (/= '=') constant))) | constant <- constantdefs]
  
  -- Outputting information about the procedure.
  when (verbose)                              $ putStrLn ("In " ++ (if ma then "MAPA " else "prCRL ") ++ "mode.\n") 
  when (verbose && (length sourcefiles) == 1) $ putStrLn ("Specification that was analysed: " ++ (sourcefiles!!0))
  when (verbose && pa) $ putStrLn ("Confluence reduction: " ++ show confluence)
  when (verbose)       $ putStrLn ("Basic reductions: " ++ show (not nobasics))
  when (verbose)       $ putStrLn ("Dead variable reduction: " ++ show dead ++ "\n")

  -- Parsing the input, and doing all the relevant transformations for prCRL mode.
  let prismTranslation                       = if fromprism then fromPrism else id
  let (basicSpec, actiontypes, untilformula) = (parseInput merging fromprism constants . prismTranslation) input
  let standardTransformations                = if nobasics then id else simplify . sumelm . simplify . constelm . sumelm
  let transformations                        = standardTransformations . (if merging then encode . mergeTauSummands . decode else id) . 
                                                (if dead                 then (constelm . deadVariableReduction)
	                                            else if includingParelm then (constelm . parelm) 
		                                        else id) . 
		                                        standardTransformations
  let specification                          = transformations basicSpec
  let lppe                                   = getLPPE specification
  --putStrLn (prismTranslation input)
 
  -- Showing the LPPE and the number of parameters / summands.
  when (pa && prcrl)                   $ print (LPPEShowPRCRL specification)
  when (pa && printLPPE && not(prcrl)) $ print (LPPEShow specification pastyle) 
  when (pa && (printLPPE || prcrl) && verbose)    $ putStrLn ("Number of parameters: " ++ show (length (getLPPEPars lppe)))
  when (pa && (printLPPE || prcrl) && verbose)    $ putStrLn ("Number of summands:   " ++ show (length (getPSummands lppe)))

  -- Detecting confluence.
  let confluents_                    = if confluence then confluentSummands specification strong else []
  let confluentSummands              = if not(removecycles) then confluents_ else removeConfluentCycles specification confluents_ 
  when (pa && confluence && verbose) $ putStrLn ("Confluent summands:   " ++ show confluentSummands)

  -- Computing the actual state space.
  let statespacesize = getStatesAndTransitionsConfluence ignorecycles specification confluentSummands True dtmc False False

  -- Showing information about the state space.
  when (pa && stateSpaceSize) $ putStrLn ("Number of states: " ++ show (fst (fst statespacesize)))
  when (pa && stateSpaceSize) $ putStrLn ("Number of transitions: " ++ show (snd (fst statespacesize)))
  when (pa && visited)        $ putStrLn ("Number of states visited: " ++ show (fst (snd statespacesize)))
  when (pa && visited)        $ putStrLn ("Number of transitions visited: " ++ show (snd (snd statespacesize)))

  -- Printing the state space to an AUT file (either probabilistically or non-probabilistically).
  when (pa && printAUTfile && dtmc)                        $ print (toAUTDTMC    ignorecycles specification cadp confluentSummands True False False)
  when (pa && printAUTfile && noprobs && not(dtmc))        $ print (toAUTNonProb ignorecycles specification cadp confluentSummands dtmc False False)
  when (pa && printAUTfile && (not(noprobs)) && not(dtmc)) $ print (toAUTProb    ignorecycles specification cadp confluentSummands dtmc False False)

  -- Printing the state space to PRISM's trans format.
  when (pa && trans)               $ putStrLn (toTrans ignorecycles specification confluentSummands dtmc False False)

  -- Printing all the actual states and transitions.
  when (pa && states) $ print (fst3 (getStateSpace ignorecycles specification confluentSummands False dtmc False False)) 
  
  -- Outputting to PRISM.
  let specificationPRISM = if confluence then doSyntacticConfluence specification confluentSummands else specification
  when (pa && prism) $ print (toPRISM actiontypes checkuntil untilformula specificationPRISM)

  -- when prism $ print (toPRISM actiontypes checkuntil untilformula specificationPRISM)
  -- when markov $ print (getCTMC mspecification)
  -- when toAUTfileMA $ print (toAUTMA ignorecycles specification cadp confluents)
 
  -- Dealing with MAPA
  let basicSpecM               = decode basicSpec
  let standardTransformationsM = if nobasics then id else decode . simplify . encode . sumelmM . decode . simplify . constelm . encode . sumelmM
  let transformationsM         = standardTransformationsM . 
	                             (if merging then mergeTauSummands else id) . 
                                 (if dead then (decode . constelm . deadVariableReduction . encode) else 
                                  if includingParelm then decode . parelm . encode else id) . 
                                 (if maxprogress then removeUnnecessaryRates else id) . 
                                  standardTransformationsM
  let specificationM           = transformationsM basicSpecM

  when (ma && verbose)                 $ putStrLn ("Number of parameters: " ++ show (getNrParamsM specificationM))
  when (ma && verbose)                 $ putStrLn ("MLPPE size: " ++ show (getMLPPESize specificationM))

  when (ma && mapa)                    $ print (MLPPEShowMAPA specificationM)
  when (ma && printMLPPE && not(mapa)) $ print (MLPPEShow specificationM pastyle) 

  let removeRates     = not keepRates
  let statespacesizeM = getStatesAndTransitionsConfluence ignorecycles (encode specificationM) confluentSummands True dtmc True removeRates

  when (ma && stateSpaceSize) $ putStrLn ("Number of states: " ++ show (fst (fst statespacesizeM)))
  when (ma && stateSpaceSize) $ putStrLn ("Number of transitions: " ++ show (snd (fst statespacesizeM)))

  when (ma && printAUTfile && noprobs)        $ print (toAUTNonProb ignorecycles (encode specificationM) cadp confluentSummands False True removeRates)
  when (ma && printAUTfile && (not(noprobs))) $ print (toAUTProb    ignorecycles (encode specificationM) cadp confluentSummands False True removeRates)

  when (ma && states) $ print (fst3 (getStateSpace ignorecycles (encode specificationM) confluentSummands False False True removeRates)) 
